/*
 * markers.c
 * kirk johnson
 * august 1995
 *
 * RCS $Id: markers.c,v 1.6 1995/09/24 00:45:16 tuna Exp $
 *
 * Copyright (C) 1989, 1990, 1993, 1994, 1995 Kirk Lauritz Johnson
 *
 * Parts of the source code (as marked) are:
 *   Copyright (C) 1989, 1990, 1991 by Jim Frost
 *   Copyright (C) 1992 by Jamie Zawinski <jwz@lucid.com>
 *
 * Permission to use, copy, modify and freely distribute xearth for
 * non-commercial and not-for-profit purposes is hereby granted
 * without fee, provided that both the above copyright notice and this
 * permission notice appear in all copies and in supporting
 * documentation.
 *
 * Unisys Corporation holds worldwide patent rights on the Lempel Zev
 * Welch (LZW) compression technique employed in the CompuServe GIF
 * image file format as well as in other formats. Unisys has made it
 * clear, however, that it does not require licensing or fees to be
 * paid for freely distributed, non-commercial applications (such as
 * xearth) that employ LZW/GIF technology. Those wishing further
 * information about licensing the LZW patent should contact Unisys
 * directly at (lzw_info@unisys.com) or by writing to
 *
 *   Unisys Corporation
 *   Welch Licensing Department
 *   M/S-C1SW19
 *   P.O. Box 500
 *   Blue Bell, PA 19424
 *
 * The author makes no representations about the suitability of this
 * software for any purpose. It is provided "as is" without express or
 * implied warranty.
 *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
 * OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "xearth.h"
#include "kljcpyrt.h"

#define BuiltInName "built-in"
#define MaxLineLen  (1024)

static void        builtin_marker_info _P((void));
static void        user_marker_info _P((char *));
static void        process_line _P((char *, const char *, int));
static const char *decode_marker_info _P((int, char **));
static const char *decode_extra_marker_info _P((int, char **, MarkerInfo *));
static char       *get_line _P((FILE *, char *, int, const char **));

MarkerInfo   *marker_info;
static ExtArr info = NULL;

/* builtin_marker_data[] contains the "built-in" marker data that is
 * compiled into xearth. (My apologies for misspellings, omissions of
 * your favorite location, or geographic inaccuracies. This is
 * primarily just a pile of data that I had handy instead of an
 * attempt to provide a sample that is "globally correct" in some
 * sense.)
 */

static const char *builtin_marker_data[] =
{
  "  61.22 -149.88 \"Anchorage\"           # Alaska, USA           ",
  "  37.97   23.72 \"Athens\"              # Greece                ",
  "  33.35   44.42 \"Baghdad\"             # Iraq                  ",
  "  13.75  100.52 \"Bangkok\"             # Thailand              ",
  "  39.92  116.42 \"Beijing\"             # China                 ",
  "  52.52   13.40 \"Berlin\"              # Germany               ",
  "  32.33  -64.75 \"Bermuda\"             # Bermuda               ",
  "  42.35  -71.07 \"Boston\"              # Massachusetts, USA    ",
  " -15.78  -47.92 \"Brasilia\"            # Brazil                ",
  "  -4.26   15.28 \"Brazzaville\"         # Congo                 ",
  " -34.60  -58.45 \"Buenos Aires\"        # Argentina             ",
  "  30.05   31.25 \"Cairo\"               # Egypt                 ",
  "  22.53   88.37 \"Calcutta\"            # India                 ",
  " -33.92   18.37 \"Cape Town\"           # South Africa          ",
  "  33.65   -7.58 \"Casablanca\"          # Morocco (Rabat?)      ",
  "  41.88  -87.63 \"Chicago\"             # Illinois, USA         ",
  "  32.78  -96.80 \"Dallas\"              # Texas, USA            ",
  "  39.72 -105.02 \"Denver\"              # Colorado, USA         ",
  "  25.30   55.30 \"Dubai\"               # UAE (Abu Dhabi?)      ",
  " -27.13 -109.38 \"Easter Island\"       # Easter Island         ",
  " -18.00  178.00 \"Fiji\"                # Fiji                  ",
  "  13.47  144.78 \"Guam\"                # Guam                  ",
  "  60.17   24.97 \"Helsinki\"            # Finland               ",
  "  10.75  106.67 \"Ho Chi Minh City\"    # Vietnam (Hanoi?)      ",
  "  22.25  114.18 \"Hong Kong\"           # Hong Kong             ",
  "  21.32 -157.86 \"Honolulu\"            # Hawaii, USA           ",
  "  52.27  104.33 \"Irkutsk\"             # Irkutsk, Russia       ",
  "  41.02   28.97 \"Istanbul\"            # Turkey (Ankara?)      ",
  "  -6.17  106.80 \"Jakarta\"             # Indonesia             ",
  "  31.76   35.23 \"Jerusalem\"           # Israel                ",
  "  34.52   69.20 \"Kabul\"               # Afghanistan           ",
  "  27.72   85.32 \"Katmandu\"            # Nepal                 ",
  "  50.43   30.52 \"Kiev\"                # Ukraine               ",
  "   3.15  101.72 \"Kuala Lumpur\"        # Malaysia              ",
  "   6.45    3.40 \"Lagos\"               # Nigeria               ",
  " -12.05  -77.05 \"Lima\"                # Peru                  ",
  "  51.50   -0.17 \"London\"              # United Kingdom        ",
  "  40.40   -3.68 \"Madrid\"              # Spain                 ",
  "  14.58  120.98 \"Manila\"              # The Phillipines       ",
  "  21.45   39.82 \"Mecca\"               # Saudi Arabia          ",
  "  19.40  -99.15 \"Mexico City\"         # Mexico                ",
  "  25.77  -80.20 \"Miami\"               # Florida, USA          ",
  "   6.30  -10.78 \"Monrovia\"            # Liberia               ",
  "  45.55  -73.57 \"Montreal\"            # Quebec, Canada        ",
  "  55.75   37.58 \"Moscow\"              # Russia                ",
  "  -1.28   36.82 \"Nairobi\"             # Kenya                 ",
  "  28.60   77.20 \"New Delhi\"           # India                 ",
  "  59.92   10.75 \"Oslo\"                # Norway                ",
  "  48.86    2.33 \"Paris\"               # France                ",
  " -31.93  115.83 \"Perth\"               # Australia             ",
  "  45.55 -122.60 \"Portland\"            # Oregon, USA           ",
  "  -0.22  -78.50 \"Quito\"               # Ecuador               ",
  "  64.15  -21.85 \"Reykjavik\"           # Iceland               ",
  " -22.90  -43.25 \"Rio de Janeiro\"      # Brazil                ",
  "  41.90   12.48 \"Rome\"                # Italy                 ",
  "  37.80 -122.40 \"San Francisco\"       # California, USA       ",
  "   9.93  -84.08 \"San Jose\"            # Costa Rica            ",
  "  18.46  -66.12 \"San Juan\"            # Puerto Rico           ",
  " -33.45  -70.67 \"Santiago\"            # Chile                 ",
  "   1.28  103.85 \"Singapore\"           # Singapore             ",
  "  42.68   23.32 \"Sofia\"               # Bulgaria              ",
  "  59.33   18.05 \"Stockholm\"           # Sweden                ",
  " -33.86  151.22 \"Sydney\"              # Australia             ",
  " -17.62 -149.45 \"Tahiti\"              # Tahiti                ",
  "  16.77   -3.02 \"Timbuktu\"            # Mali (Bamako?)        ",
  "  35.70  139.77 \"Tokyo\"               # Japan                 ",
  "  43.65  -79.38 \"Toronto\"             # Ontario, Canada       ",
  "  32.90   13.18 \"Tripoli\"             # Libya                 ",
  "  47.92  106.88 \"Ulan Bator\"          # Mongolia              ",
  "  49.27 -123.12 \"Vancouver\"           # B.C., Canada          ",
  "  48.22   16.33 \"Vienna\"              # Austria               ",
  "  38.90  -77.02 \"Washington\"          # United States         ",
  " -41.30  174.77 \"Wellington\"          # New Zealand           ",
  "  62.45 -114.35 \"Yellowknife\"         # N.T., Canada          ",
  "  90.00    0.00 \"North Pole\"          # North Pole            ",
  " -90.00    0.00 \"South Pole\"          # South Pole            ",
  NULL
};


void load_marker_info(fname)
     char *fname;
{
  int         i;
  MarkerInfo *new;

  if (info == NULL)
  {
    /* first time through, allocate a new extarr
     */
    info = extarr_alloc(sizeof(MarkerInfo));
  }
  else
  {
    /* on subsequent passes, just clean it up for reuse
     */
    for (i=0; i<(info->count-1); i++)
      free(((MarkerInfo *) info->body)[i].label);
    info->count = 0;
  }

  if (strcmp(fname, BuiltInName) == 0)
    builtin_marker_info();
  else
    user_marker_info(fname);

  new = (MarkerInfo *) extarr_next(info);
  new->lat   = 0;
  new->lon   = 0;
  new->label = NULL;

  marker_info = (MarkerInfo *) info->body;
}


static void builtin_marker_info()
{
  int   idx;
  char *buf;

  buf = (char *) malloc(MaxLineLen);
  assert(buf != NULL);

  idx = 0;
  while (builtin_marker_data[idx] != NULL)
  {
    strcpy(buf, builtin_marker_data[idx]);
    idx += 1;
    process_line(buf, BuiltInName, idx);
  }

  free(buf);
}


static void user_marker_info(fname)
     char *fname;
{
  int         idx;
  char       *buf;
  const char *msg;
  FILE       *ins;

  ins = fopen(fname, "r");
  if (ins == NULL)
  {
    warning("unable to open marker info file");
    return;
  }

  idx = 1;
  buf = (char *) malloc(MaxLineLen);
  assert(buf != NULL);

  while (get_line(ins, buf, MaxLineLen, &msg) != NULL)
  {
    if (msg != NULL)
    {
      fflush(stdout);
      fprintf(stderr, "%s, line %d: %s\n", fname, idx, msg);
      fprintf(stderr, " (length = %d)\n", (int) strlen(buf));
      fflush(stderr);
    }

    process_line(buf, fname, idx);
    idx += 1;
  }

  free(buf);
  fclose(ins);
}


static void process_line(buf, fname, idx)
     char       *buf;
     const char *fname;
     int         idx;
{
  int         tokc;
  char      **tokv;
  const char *msg;

  tokv = tokenize(buf, &tokc, &msg);
  if (msg != NULL)
  {
    fflush(stdout);
    fprintf(stderr, "%s, line %d: %s\n", fname, idx, msg);
    fflush(stderr);
  }

  msg = decode_marker_info(tokc, tokv);
  if (msg != NULL)
  {
    fflush(stdout);
    fprintf(stderr, "%s, line %d: %s\n", fname, idx, msg);
    fflush(stderr);
  }

  free(tokv);
}



static const char *decode_marker_info(tokc, tokv)
     int    tokc;
     char **tokv;
{
  double      lat;
  double      lon;
  char       *label;
  MarkerInfo *new;
  const char *rslt;

  if (tokc == 0)
  {
    /* blank or comment line, no problem
     */
    rslt = NULL;
  }
  else if (tokc < 3)
  {
    /* less than three tokens is a problem
     */
    rslt = "syntax error in required info";
  }
  else
  {
    /* decode first three tokens
     */
    sscanf(tokv[0], "%lf", &lat);
    sscanf(tokv[1], "%lf", &lon);
    label = (char *) malloc(strlen(tokv[2]) + 1);
    assert(label != NULL);
    strcpy(label, tokv[2]);

    if ((lat > 90) || (lat < -90))
    {
      rslt = "latitude must be between -90 and 90";
    }
    else if ((lon > 180) || (lon < -180))
    {
      rslt = "longitude must be between -180 and 180";
    }
    else
    {
      new = (MarkerInfo *) extarr_next(info);
      new->lat   = lat;
      new->lon   = lon;
      new->label = label;

      /* decode any extra tokens
       */
      rslt = decode_extra_marker_info(tokc-3, tokv+3, new);
    }
  }

  return rslt;
}


static const char *decode_extra_marker_info(tokc, tokv, inf)
     int        tokc;
     char     **tokv;
     MarkerInfo *inf;
{
  int   i;
  char *key;
  char *val;

  /* set defaults
   */
  inf->align = MarkerAlignDefault;

  for (i=0; i<tokc; i++)
  {
    /* look for '=' separator; quit gracefully if none found, else
     * change it to a NUL and set key and val appropriately
     */
    val = tokv[i];
    while ((*val != '=') && (*val != '\0'))
      val += 1;

    if (*val == '\0')
      return "syntax error in optional info";

    key  = tokv[i];
    *val = '\0';
    val += 1;

    /* big decoding switch
     */
    if (strcmp(key, "align") == 0)
    {
      if (strcmp(val, "left") == 0)
        inf->align = MarkerAlignLeft;
      else if (strcmp(val, "right") == 0)
        inf->align = MarkerAlignRight;
      else if (strcmp(val, "above") == 0)
        inf->align = MarkerAlignAbove;
      else if (strcmp(val, "below") == 0)
        inf->align = MarkerAlignBelow;
      else
        return "unrecognized alignment value";
    }
    else
    {
      return "unrecognized key in optional info";
    }
  }

  return NULL;
}


static char *get_line(ins, buf, len, msg)
     FILE        *ins;
     char        *buf;
     int          len;
     const char **msg;
{
  int   ch;
  char *rslt;

  *msg = NULL;
  rslt = buf;

  while (1)
  {
    ch = getc(ins);

    if (ch == EOF)
    {
      if (buf == rslt)
      {
        /* if EOF is the first thing we read on a line,
         * there aren't any more lines to read
         */
        rslt = NULL;
        break;
      }
      else
      {
        /* line terminated by EOF
         */
        *buf = '\0';
        break;
      }
    }
    else if (ch == '\n')
    {
      /* line terminated by newline
       */
      *buf = '\0';
      break;
    }
    else
    {
      *buf = ch;
      buf += 1;
      len -= 1;

      if (len == 0)
      {
        /* line terminated by exceeding maximum length
         */
        *msg = "line exceeds maximum length";
        *(buf-1) = '\0';
        do
          ch = getc(ins);
        while ((ch != EOF) && (ch != '\n'));
        break;
      }
    }
  }

  return rslt;
}


void show_marker_info(fname)
     char *fname;
{
  MarkerInfo *minfo;

  printf("# xearth marker data from \"%s\":\n", fname);

  load_marker_info(fname);

  minfo = marker_info;
  while (minfo->label != NULL)
  {
    printf("%7.2f %8.2f \"%s\"", minfo->lat, minfo->lon, minfo->label);

    if (minfo->align == MarkerAlignLeft)
      printf(" align=left");
    else if (minfo->align == MarkerAlignRight)
      printf(" align=right");
    else if (minfo->align == MarkerAlignAbove)
      printf(" align=above");
    else if (minfo->align == MarkerAlignBelow)
      printf(" align=below");
    else
      assert(minfo->align == MarkerAlignDefault);

    printf("\n");
    minfo += 1;
  }

  exit(0);
}
